@@ -4,6 +4,7 @@ from .tabulate import tabulate as tabulate_
 import sys
 import pandas as pd
 import re
+import datetime
 
 
 def _get_version():
@@ -35,6 +36,7 @@ A list of available commands follows.
 """
 
 COMMANDS = {}
+DOCS = {}
 
 
 def _gpx(fname):
@@ -70,13 +72,16 @@ READERS = {
     "fwf": pd.read_fwf,
     "json": pd.read_json,
     "html": pd.read_html,
-    "excel": lambda fname, sheet: pd.read_excel(fname, sheet),
-    "xls": lambda fname, sheet: pd.read_excel(fname, sheet),
-    "odf": lambda fname, sheet: pd.read_excel(fname, sheet),
     "tsv": _tsv,
     "gpx": _gpx,
 }
 
+try:
+    READERS["excel"] = pd.read_excel
+    READERS["xls"] = pd.read_excel
+    READERS["odf"] = pd.read_excel
+except AttributeError:
+    pass
 
 try:
     READERS["hdf5"] = pd.read_hdf
@@ -179,10 +184,21 @@ FALSY = ("False", "false", "No", "no", "0", False, 0, "None")
 TRUTHY = ("True", "true", "Yes", "yes", "1", True, 1)
 
 
+def _assert_col(df, col):
+    if col not in df.columns:
+        exit("Unknown column {}".format(col))
+
+
+def _assert_cols(df, cols):
+    for col in cols:
+        _assert_col(df, col)
+
+
 def register(fn, name=None):
     if name is None:
         name = fn.__name__
     COMMANDS[name] = fn
+    DOCS[name] = fn.__doc__
     return fn
 
 
@@ -250,6 +266,18 @@ def dataset(dset=None):
     pipeout(df)
 
 
+@register
+def diff(*cols, periods=1, axis=0):
+    df = pipein()
+    if not cols:
+        df = df.diff(periods=periods, axis=axis)
+    else:
+        _assert_cols(df, cols)
+        columns = list(cols)
+        df[columns] = df[columns].diff(periods=periods, axis=axis)
+    pipeout(df)
+
+
 @register
 def dropna(axis=0, how="any", thresh=None):
     """Remove rows (or columns) with N/A values.
@@ -366,7 +394,7 @@ def fillna(value=None, method=None, limit=None):
         value = __tryparse(value)
         pipeout(pipein().fillna(value=value, limit=limit))
     else:
-        exit("'ph dropna' needs exactly one of value and method")
+        exit("'ph fillna' needs exactly one of value and method")
 
 
 @register
@@ -397,12 +425,36 @@ def appendstr(col, s, newcol=None):
 @register
 def split(col, pat=" "):
     df = pipein()
-    if col not in df.columns:
-        exit("No such column {}".format(col))
+    _assert_col(df, col)
     df[[col, col + "_rhs"]] = df[col].str.split(pat=pat, n=1, expand=True)
     pipeout(df)
 
 
+@register
+def strip(*cols, lstrip=False, rstrip=False):
+    """Strip (trim) a string.
+
+    Usage: cat x.csv | ph strip
+           cat x.csv | ph strip --lstrip=True
+           cat x.csv | ph strip --rstrip=True
+
+    """
+    df = pipein()
+    if not cols:
+        cols = list(df.columns)
+    else:
+        cols = list(cols)
+    _assert_cols(df, cols)
+    for c in cols:
+        if lstrip in TRUTHY:
+            df[c] = df[c].str.lstrip()
+        elif rstrip in TRUTHY:
+            df[c] = df[c].str.rstrip()
+        else:
+            df[c] = df[c].str.strip()
+    pipeout(df)
+
+
 @register
 def astype(type, column=None, newcolumn=None):
     """Cast a column to a different type.
@@ -496,9 +548,7 @@ def groupby(*columns, how="sum", as_index=False):
     if not columns:
         exit("Needs at least one column to group by")
     df = pipein()
-    for c in columns:
-        if c not in df.columns:
-            exit("Unknown column name {}".format(c))
+    _assert_cols(df, columns)
     if as_index in TRUTHY:
         as_index = True
     elif as_index in FALSY:
@@ -517,15 +567,26 @@ def groupby(*columns, how="sum", as_index=False):
 
 
 @register
-def rolling(window, how="sum", win_type=None, std=None, beta=None, tau=None):
+def rolling(window, *columns, how="sum", win_type=None, std=None, beta=None, tau=None):
     """Rolling window calculations using provided `how` function.
 
     Usage: cat a.csv | ph rolling 3
+           cat a.csv | ph rolling 5 --how=mean
+           cat a.csv | ph rolling 5 colA colB --how=mean
            cat a.csv | ph rolling 5 --win_type=gaussian --std=7.62
     """
     df = pipein()
+    orig_columns = list(df.columns)
+    columns = list(columns)
+    _assert_cols(df, columns)
 
-    rollin = df.rolling(window, win_type=win_type)
+    if not columns:
+        columns = list(df.columns)
+
+    noncols = [c for c in df.columns if c not in columns]
+
+    rollin = df[columns].rolling(window, win_type=win_type)
+    nonrollin = df[noncols]
     try:
         fn = getattr(rollin, how)
     except AttributeError:
@@ -536,7 +597,9 @@ def rolling(window, how="sum", win_type=None, std=None, beta=None, tau=None):
     else:
         retval = fn()
 
-    pipeout(retval)
+    df = pd.concat([retval, nonrollin], axis=1)
+    df = df[orig_columns]
+    pipeout(df)
 
 
 @register
@@ -716,19 +779,12 @@ def plot(*args, **kwargs):
     index = kwargs.get("index")
     if index is not None:
         df = df.set_index(index)
+        del kwargs["index"]
     for log_ in ("logx", "logy", "loglog"):
         if kwargs.get(log_) in TRUTHY:
             kwargs[log_] = True
 
-    df.plot(
-        kind=kwargs.get("kind", "line"),  # default pandas plot is line
-        style=kwargs.get("style"),
-        x=kwargs.get("x"),
-        y=kwargs.get("y"),
-        logx=kwargs.get("logx"),
-        logy=kwargs.get("logy"),
-        loglog=kwargs.get("loglog"),
-    )
+    df.plot(**kwargs)
     plt.show()
     pipeout(df)
 
@@ -764,7 +820,7 @@ def normalize(col=None):
 
 
 @register
-def date(col=None, unit=None, origin="unix", errors="raise", dayfirst=False):
+def date(col=None, unit=None, origin="unix", errors="raise", dayfirst=False, **kwargs):
     """Assemble datetime from multiple columns or from one column
 
     --unit can be D, s, us, ns (defaults to ns, ns from origin)
@@ -773,10 +829,16 @@ def date(col=None, unit=None, origin="unix", errors="raise", dayfirst=False):
 
     --errors can be raise, coerce, ignore (see pandas.to_datetime)
 
+    --format a strptime format string, e.g. '%Y-%m-%d %H:%M:%S'
+
+    --utc=True if the input is in utc, i.e. seconds from epoch
+
     Usage: cat a.csv | ph date x
            cat a.csv | ph date x --unit=s --origin="1984-05-17 09:30"
            cat a.csv | ph date x --dayfirst=True
            cat a.csv | ph date  # if a.csv contains year, month, date
+           cat a.csv | ph date x --format="%Y-%m-%d"
+           cat a.csv | ph date x --utc=True
 
     """
     DATE_ERRORS = ("ignore", "raise", "coerce")
@@ -785,16 +847,25 @@ def date(col=None, unit=None, origin="unix", errors="raise", dayfirst=False):
 
     dayfirst = dayfirst in TRUTHY
 
+    date_parser = None
+    if "format" in kwargs:
+        date_parser = lambda d: [
+            datetime.datetime.strptime(str(e), kwargs["format"]) for e in d
+        ]
+    if kwargs.get("utc") in TRUTHY:
+        date_parser = lambda d: [datetime.datetime.utcfromtimestamp(e) for e in d]
     df = pipein()
     try:
         if col is None:
             df = pd.to_datetime(df, unit=unit, origin=origin, errors=errors)
         else:
-            if col not in df.columns:
-                exit("No such column {}".format(col))
-            df[col] = pd.to_datetime(
-                df[col], unit=unit, origin=origin, errors=errors, dayfirst=dayfirst
-            )
+            _assert_col(df, col)
+            if date_parser is None:
+                df[col] = pd.to_datetime(
+                    df[col], unit=unit, origin=origin, errors=errors, dayfirst=dayfirst
+                )
+            else:
+                df[col] = date_parser(df[col])
     except Exception as err:
         exit(err)
 
@@ -1033,8 +1104,8 @@ def _print_commands(cmds):
     print(tabulate_(df.transpose(), showindex=False))
 
 
-@register
-def help(*args, **kwargs):
+@registerx("help")
+def help_(*args, **kwargs):
     """Writes help (docstring) about the different commands."""
     if not args:
         print("Usage: ph command arguments")
@@ -1042,23 +1113,35 @@ def help(*args, **kwargs):
         _print_commands(sorted(COMMANDS.keys()))
         exit(0)
     cmd = args[0]
-    import ph
 
-    try:
-        fn = getattr(ph, cmd)
-        ds = getattr(fn, "__doc__")
-    except AttributeError:
+    ds = None
+    if cmd in DOCS:
+        ds = DOCS[cmd]
+    else:
         try:
             fn = getattr(pd.DataFrame, cmd)
             ds = getattr(fn, "__doc__")
         except AttributeError:
-            ds = ""
-
-    print("Usage: ph {} [?]".format(cmd))
-    print("       {}".format(ds))
+            pass
+    if ds is None:
+        exit("Unknown command {}".format(cmd))
+    print("Usage: ph {}".format(cmd))
+    print("       {}".format(ds.strip()))
 
 
 def slugify_name(name):
+    name_ = name
+    try:
+        name = float(name_)
+    except ValueError:
+        pass
+    if isinstance(name_, (int, str)):
+        try:
+            name = int(name_)
+        except ValueError:
+            pass
+    if isinstance(name, (int, float)):
+        name = str(name) + "_"
     if not name:
         return "unnamed"
     if name == "_":
@@ -1099,6 +1182,21 @@ def slugify():
     pipeout(df)
 
 
+@register
+def raw(fname=None):
+    """Do your best to read this comma-separated input."""
+    import csv
+
+    if fname is None:
+        d = csv.reader(sys.stdin)
+        df = pd.DataFrame(d)
+    else:
+        with open(fname, "r") as fin:
+            d = csv.reader(fin)
+            df = pd.DataFrame(d)
+    pipeout(df)
+
+
 @registerx("open")
 def open_(ftype, fname, **kwargs):
     """Use a reader to open a file.
@@ -1112,8 +1210,8 @@ def open_(ftype, fname, **kwargs):
            ph open excel a.ods
            ph open excel a.xls
            ph open excel a.xlsx
-           ph open excel a.xls --sheet=2
-           ph open excel a.xls --sheet="The Real Dataset sheet"
+           ph open excel a.xls --sheet_name=2
+           ph open excel a.xls --sheet_name="The Real Dataset sheet"
            ph open csv a.csv --thousands=','
 
 
@@ -1128,10 +1226,6 @@ def open_(ftype, fname, **kwargs):
         exit("Unknown filetype {}".format(ftype))
     reader = READERS[ftype]
 
-    sheet = kwargs.get("sheet")
-    if ftype in ("excel", "xls", "odf"):
-        kwargs["sheet"] = __tryparse(sheet)
-
     if kwargs.get("sep") == "\\t":
         kwargs["sep"] = "\t"
 
@@ -1153,6 +1247,19 @@ def open_(ftype, fname, **kwargs):
     try:
         if ftype == "clipboard":
             df = reader(**kwargs)
+        elif ftype in ("excel", "xls", "odf"):
+            try:
+                df = reader(fname, **kwargs)
+            except Exception as err:
+                exit(err)
+            if not isinstance(df, pd.DataFrame):  # could be dict
+                try:
+                    errormsg = 'Specify --sheet_name="{}"'.format(
+                        "|".join(str(k) for k in df.keys())
+                    )
+                except Exception:
+                    errormsg = "Specify --sheet_name"
+                exit(errormsg)
         else:
             df = reader(fname, **kwargs)
     except AttributeError as err:
@@ -1315,9 +1422,7 @@ def columns(*cols, **kwargs):
             if col.endswith(q) and col not in cols:
                 cols.append(col)
 
-    for col in cols:
-        if col not in df.columns:
-            exit("No such column {}".format(col))
+    _assert_cols(df, cols)
 
     if not cols and not kwargs:
         print("columns")
@@ -1326,6 +1431,84 @@ def columns(*cols, **kwargs):
         pipeout(df[cols])
 
 
+@register
+def spencer(*cols):
+    """Compute Spencer's 15-weight average.
+
+    Usage: cat a.csv | ph spencer
+
+    Experimental feature for computing Spencer's 15-weight average.
+    Smooths out curves by removing high frequency noise.  Will
+    ultimately lose some data on each end of the timeseries.
+
+    """
+    _SPENCER = (-3, -6, -5, 3, 21, 46, 67, 74, 67, 46, 21, 3, -5, -6, -3)
+    _SPENCER_SUM = sum(_SPENCER)
+
+    def spencer_(lst):
+        for i in range(7, len(lst) - 8):
+            seq = lst[i - 7 : i + 8]
+            yield sum(seq[i] * _SPENCER[i] / _SPENCER_SUM for i in range(15))
+
+    df = pipein()
+    _assert_cols(df, cols)
+    prefix = [float("nan")] * 7
+    suffix = [float("nan")] * 8
+    if not cols:
+        cols = list(df.columns)
+    for col in cols:
+        lst = list(df[col])
+        s = list(spencer_(lst))
+        ncol = prefix + s + suffix
+        df[col] = ncol
+    pipeout(df)
+
+
+def _parse_slice(slicestr):
+    pattern = ":<int> | <int>: | <int>:<int> | <int>:<int>:<int>"
+    error = "Input to slice is {} _not_ {}".format(pattern, slicestr)
+
+    assert ":" in slicestr, error
+    start = None
+    end = None
+    step = None
+    tup = slicestr.split(":")
+    if len(tup) > 3:
+        exit(error)
+    start = tup[0] or None
+    if start is not None:
+        start = int(start)
+    end = tup[1] or None
+    if end is not None:
+        end = int(end)
+    if len(tup) == 3:
+        step = tup[2] or None
+        if step is not None:
+            step = int(step)
+    return start, end, step
+
+
+@registerx("slice")
+def slice_(slicestr):
+    """Slice a dataframe with Python slice pattern.
+
+    Usage: cat a.csv | ph slice :10    # head
+           cat a.csv | ph slice -10:   # tail
+           cat a.csv | ph slice ::2    # every even row
+           cat a.csv | ph slice 1::2   # every odd row
+           cat a.csv | ph slice ::-1   # reverse file
+
+    """
+    pattern = ":<int> | <int>: | <int>:<int> | <int>:<int>:<int>"
+    error = "Input to slice is {} _not_ {}".format(pattern, slicestr)
+    df = pipein()
+    if isinstance(slicestr, int) or ":" not in slicestr:
+        exit(error)
+    start, end, step = _parse_slice(slicestr)
+    retval = df[start:end:step]
+    pipeout(retval)
+
+
 @register
 def drop(*columns, **kwargs):
     """Drop specified labels from rows or columns.
@@ -1348,9 +1531,7 @@ def drop(*columns, **kwargs):
     if kwargs.get("axis") in (None, 0, "index"):
         columns = [__tryparse(col) for col in columns]
     elif kwargs.get("axis") in (1, "columns"):
-        for col in columns:
-            if col not in df.columns:
-                exit("Unknown column {}.".format(col))
+        _assert_cols(df, columns)
     else:
         exit(
             "--axis=index (or 0) or --axis=columns (or 1), not {}".format(
@@ -1401,8 +1582,7 @@ def sort(col):
 
     """
     df = pipein()
-    if col not in df.columns:
-        exit("Unknown column {}".format(col))
+    _assert_col(df, col)
     pipeout(df.sort_values(col))
 
 
@@ -1418,10 +1598,7 @@ def polyfit(x, y, deg=1):
 
     """
     df = pipein()
-    if x not in df.columns:
-        exit("Unknown column x={}".format(x))
-    if y not in df.columns:
-        exit("Unknown column y={}".format(y))
+    _assert_cols(df, (x, y))
     deg = __tryparse(deg)
     if not isinstance(deg, int) or deg <= 0:
         exit("deg={} should be a positive int".format(deg))
@@ -1443,6 +1620,8 @@ def __process(attr):
         return False
     if attr.startswith("to_"):
         return False
+    if attr == "T":
+        return False
     return True
 
 
@@ -1458,6 +1637,8 @@ def _main(argv):
     if cmd in ("-v", "--version"):
         print_version()
         exit()
+    if cmd in ("--h", "--help"):
+        cmd = "help"
     if cmd not in COMMANDS:
         exit("Unknown command {}.".format(cmd))
 
